use std::{
    env,
    error::Error,
    fmt, fs,
    io::{Cursor, Read},
    path::PathBuf,
};

use object::{Object, ObjectSection, read::pe::PeFile64};

type Result<T> = std::result::Result<T, Box<dyn Error>>;

#[derive(Debug)]
enum MyError {
    NoParentDir,
    NoSection,
    NoData,
}
impl fmt::Display for MyError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::NoParentDir => write!(f, "no parent dir for path"),
            Self::NoSection => write!(f, "no section found"),
            Self::NoData => write!(f, "no data"),
        }
    }
}
impl Error for MyError {}

fn next_u64_le(r: &mut impl Read) -> Result<u64> {
    let mut b = [0u8; 8];
    r.read_exact(&mut b)?;
    Ok(u64::from_le_bytes(b))
}

fn main() -> Result<()> {
    // read args
    let mut args = env::args().skip(1);
    let table_base: u64 = args
        .next()
        .and_then(|s| u64::from_str_radix(&s, 16).ok())
        .expect("no table base provided");
    let table_len: u64 = args
        .next()
        .and_then(|s| s.parse().ok())
        .expect("no table len provided");
    let path = args.next().expect("no *.exe provided");

    // read file
    let data = fs::read(path)?;

    // parse file
    let file = PeFile64::parse(&*data)?;

    // find section
    let section = file
        .sections()
        .find(|s| (s.address()..s.address() + s.size()).contains(&table_base))
        .ok_or(MyError::NoSection)?;

    // dump table
    for i in 0..table_len {
        // get entry
        let ptr_entry = table_base + 32 * i;
        let entry = section.data_range(ptr_entry, 32)?.ok_or(MyError::NoData)?;

        // parse entry
        let mut c = Cursor::new(entry);
        let ptr_name = next_u64_le(&mut c)?;
        let len_name = next_u64_le(&mut c)?;
        let ptr_data = next_u64_le(&mut c)?;
        let len_data = next_u64_le(&mut c)?;

        // read name/data
        let name = section
            .data_range(ptr_name, len_name)?
            .ok_or(MyError::NoData)?;
        let data = section
            .data_range(ptr_data, len_data)?
            .ok_or(MyError::NoData)?;

        // print name
        let mut name = String::from_utf8(name.to_vec())?;
        println!("saving {name}");

        // make path
        let mut p = PathBuf::from("./save");
        name.insert(0, '.');
        p.push(name);

        // open file
        let parent = p.parent().ok_or(MyError::NoParentDir)?;
        fs::create_dir_all(parent)?;
        let mut file = fs::File::create_new(p)?;
        brotli::BrotliDecompress(&mut &data[..], &mut file)?;
    }

    Ok(())
}
